import numpy as np 
import cv2
import math
import random
import warnings
from ..utils import *
from skimage.transform import resize

class Crop:
	def __call__(self, image, x, y):
		rd = resize(image, (x, y))
		return rd
		
class WindowDicom:
	def window_image(img,window_center,window_width,intercept,slope):

		img = (img*slope + intercept)
		minimg = window_center - window_width//2
		maximg = window_center + window_width//2
		img[img<minimg] = minimg
		img[img>maximg] = maximg

		img = (img - minimg) / (maximg - minimg)
		return img

	def get_first_of_dicom_field_as_int(x):
    #get x[0] as in int is x is a 'pydicom.multival.MultiValue', otherwise get int(x)
		if type(x) == pydicom.multival.MultiValue:
			return int(x[0])
		else:
			return int(x)

	def get_windowing(data):
		dicom_fields = [data[('0028','1050')].value, #window center
                    data[('0028','1051')].value, #window width
                    data[('0028','1052')].value, #intercept
                    data[('0028','1053')].value] #slope
		return [get_first_of_dicom_field_as_int(x) for x in dicom_fields]


	def __call__(dicom_file, param_1, param_2):
		try:
			window_center , window_width, intercept, slope = get_windowing(dicom_file)
			img                                            = dicom_file.pixel_array
			dls                                            = window_image(img, param_1, param_2, intercept, slope)
		except Exception:
			raise("Error with windowing the data. Please check your params and retry.")

    
transformation = Crop()
Crop = CoreTransformation(p=0.5, construct_class=transformation)
transformation = WindowDicom()
WindowDicom = CoreTransformation(p=0.5, construct_class=transformation)


from collections import OrderedDict
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

def l2_norm(input, axis=1):
    norm = torch.norm(input, 2, axis, True)
    output = torch.div(input, norm)
    return output

class FocalLoss(nn.Module):
    def __init__(self, gamma=0, alpha=None, size_average=True):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float, int)): self.alpha = torch.Tensor([alpha, 1 - alpha])
        if isinstance(alpha, list): self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        if input.dim() > 2:
            input = input.view(-1, 1)
        target = target.view(-1, 1)

        pt = torch.sigmoid(input)
        pt = 1 - (pt - target.float()).abs()
        logpt = pt.log()

        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.long().data.view(-1))
            logpt = logpt * Variable(at)

        loss = -1 * (1 - pt) ** self.gamma * logpt
        if self.size_average:
            return loss.mean()
        else:
            return loss.sum()


class SEModule(nn.Module):

    def __init__(self, channels, reduction):
        super(SEModule, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc1 = nn.Conv2d(channels, channels // reduction, kernel_size=1,
                             padding=0)
        self.relu = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv2d(channels // reduction, channels, kernel_size=1,
                             padding=0)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        module_input = x
        x = self.avg_pool(x)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)
        return module_input * x


class SoftDiceLoss(nn.Module):
    def __init__(self):
        super(SoftDiceLoss, self).__init__()

    def forward(self, logits, targets):
        smooth = 1
        num = targets.size(0)
        probs = F.sigmoid(logits)
        m1 = probs.view(num, -1)
        m2 = targets.view(num, -1)
        intersection = (m1 * m2)
        score = 2. * (intersection.sum(1) + smooth) / (m1.sum(1) + m2.sum(1) + smooth)
        score = 1 - score.sum() / num
        return score


class UpBlock(nn.Module):
    def __init__(self, in_c, mid_c, out_c):
        super(UpBlock, self).__init__()
        self.conv = nn.Sequential(
            SEModule(in_c, reduction=16),
            nn.Conv2d(in_c, mid_c, kernel_size=1),
            nn.BatchNorm2d(mid_c),
            nn.ReLU(),
            nn.UpsamplingNearest2d(scale_factor=2),
            nn.Conv2d(mid_c, mid_c, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_c),
            nn.ReLU(),
            nn.Conv2d(mid_c, out_c, kernel_size=1),
            nn.ReLU()
        )
        # todo add seblock!

    def forward(self, x):
        return self.conv(x)


class CoreTransformation:

	def __init__(self, p, construct_class):
		super(CoreTransformation, self).__init__()
		self.p = p
		self.construct_class = construct_class

	def __call__(self, *args):

		if self.construct_class == None:
			raise ValueError("Constructor improperly called, please retry by properly applying a constructor for the transformation.")
		else:
			try:
				return self.construct_class(*args)
			except Exception:
				raise ValueError("Your constructor function has an error in it.")

			
class TwoImageTransformation:

	def __init__(self, p, construct_class):
		super(TwoImageTransformation, self).__init__()
		self.p = p
		self.construct_class = construct_class

	def __call__(self, image1, image2):
		if image1 == None or image2 == None:
			raise ValueError("You have improperly called the image constructor.")
		else:
			if self.construct_class == None:
				raise ValueError("Constructor improperly called, please retry by properly applying a constructor for the transformation.")
			else:
				try:
					return self.construct_class(image1, image2)
				except Exception:
					raise ValueError("Your constructor function has an error in it.")		